package checkers.igj;

import java.lang.annotation.Annotation;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.*;

import com.sun.source.tree.*;
import com.sun.source.util.SimpleTreeVisitor;
import com.sun.source.util.TreePath;
import com.sun.tools.javac.comp.TargetType;

import checkers.igj.quals.*;
import checkers.types.*;
import checkers.util.*;
import checkers.util.GenericsUtils.LocationVisitor;

/**
 * Adds support for IGJ annotations ({@code @ReadOnly},
 * {@code @Mutable}, {@code @Immutable}) to {link
 * AnnotatedTypeFactory}. This means that the
 * {@link IGJAnnotatedTypeFactory#getClass} and
 * {@link IGJAnnotatedTypeFactory#getMethod} methods will infer IGJ
 * immutability type even if they do not carry an IGJ annotation.
 * There include:
 * 
 * <ul>
 * <li>primitive, class, and enum types
 * <li>the results of binary operations
 * <li>literals (except the null literal) and arrays of literals
 * </ul>
 */
public class IGJAnnotatedTypeFactory extends AnnotatedTypeFactory {

    /** Default IGJ Immutability Annotation for method receivers * */
    protected final static Class<? extends Annotation> DEFAULT_METHOD_RECEIVER_ANNOTATION =
        Mutable.class;

    protected final static Class<? extends Annotation> DEFAULT_CONSTRUCTOR_RECEIVER_ANNOTATION =
        AssignsFields.class;
    
    /** Default IGJ Immutability Annotation for classes and instances * */
    protected final static Class<? extends Annotation> DEFAULT_CLASS_ANNOTATION =
        Mutable.class;

    /** Default IGJ Immutability Annotation for new class trees * */
    protected final static Class<? extends Annotation> DEFAULT_NEW_CLASS_ANNOTATION =
        IGJPlaceHolder.class;

    // TODO: Refactor code
    protected VisitorState state;

    public IGJAnnotatedTypeFactory(ProcessingEnvironment env,
            CompilationUnitTree root) {
        super(env, root);
    }

    @Override
    public AnnotatedClassType getClass(Element element) {
        AnnotatedClassType type = super.getClass(element);
        
        if (element.getKind() != ElementKind.CLASS
                && element.getKind() != ElementKind.INTERFACE) {
            for (AnnotationLocation loc : 
                immutableFinder.visit(element.asType()))
                type.includeAt(Immutable.class, loc);
        } else {
            if (isImmutable(element.asType(), false))
                type.includeAt(Immutable.class, AnnotationLocation.RAW);
        }
        return type;
    }

    private ClassTypeVisitor classTypeVisitor = new ClassTypeVisitor(this);

    @Override
    public AnnotatedClassType getClass(Tree tree) {
        AnnotatedClassType type = super.getClass(tree);
        
        type = classTypeVisitor.visit(tree, type);

        // Handle primitives
        if (tree.getKind() != Tree.Kind.CLASS){
            TypeMirror typeValue = 
                trees.getTypeMirror(TreePath.getPath(root, tree));

            for (AnnotationLocation loc : 
                immutableFinder.visit(typeValue, null)) {
                if (IGJImmutability.getIGJAnnotationDataAt(type, loc) == null)
                    type.includeAt(Immutable.class, loc);
            }
        }
        

        // If we don't have it
        if (IGJImmutability.getIGJAnnotationDataAt(type, AnnotationLocation.RAW) == null)
            type.include(Mutable.class);
        
        return type;
    }

    @Override
    public AnnotatedMethodType getMethod(Tree tree) {
        if (tree.getKind() == Tree.Kind.NEW_CLASS)
            return getConstructor((NewClassTree)tree);
        
        AnnotatedMethodType method = super.getMethod(tree);

        // Resolve IGJ Annotation
        if (tree.getKind() == Tree.Kind.METHOD_INVOCATION) {
            MethodInvocationTree methodInvTree = (MethodInvocationTree) tree;
            // Resolve
            if (!ElementUtils.isStatic(method.getElement())
                    && methodInvTree.getMethodSelect().getKind() == Tree.Kind.MEMBER_SELECT) {
                MemberSelectTree selectTree =
                    (MemberSelectTree) methodInvTree.getMethodSelect();
                AnnotatedClassType enclosingClass =
                    getClass(selectTree.getExpression());
                resolveBindings(method, getBinding(enclosingClass));
            }
            List<AnnotatedClassType> arguments =
                new ArrayList<AnnotatedClassType>();
            for (Tree arg : methodInvTree.getArguments())
                arguments.add(getClass(arg));
            resolveBindings(method, arguments);
        }

        return method;
    }

    protected AnnotatedMethodType getConstructor(NewClassTree tree) {
        ExecutableElement constructor = InternalUtils.constructor(tree);
        AnnotatedClassType consEncClass = getClass(ElementUtils.enclosingClass(constructor));
        
        AnnotatedMethodType method = getMethod(constructor);
        
        // if Receiver not specified
        final AnnotatedClassType receiverType = method.getAnnotatedReceiverType();
        if (IGJImmutability.getIGJAnnotationDataAt(
                receiverType, AnnotationLocation.RAW) == null)
            receiverType.include(DEFAULT_CONSTRUCTOR_RECEIVER_ANNOTATION);
        
        // FIXME: Handle inner classes
        final AnnotatedClassType returnType = method.getAnnotatedReturnType();
        final AnnotationData properReturnAnnotation = 
            IGJImmutability.getIGJAnnotationDataAt(consEncClass, AnnotationLocation.RAW);
        if (properReturnAnnotation != null)
            returnType.include(properReturnAnnotation);
        else
            returnType.include(DEFAULT_NEW_CLASS_ANNOTATION);
        
        // Try to resolve it
        List<AnnotatedClassType> arguments =
            new ArrayList<AnnotatedClassType>();
        for (Tree arg : tree.getArguments())
            arguments.add(getClass(arg));
        resolveBindings(method, arguments);
        
        return method;
    }
    
    @Override
    protected List<AnnotatedClassType> paramTypes(ExecutableElement method) {
        List<AnnotatedClassType> paramTypes =
            new LinkedList<AnnotatedClassType>();
        for (VariableElement varElement : method.getParameters()) {
            paramTypes.add(getClass(varElement));
        }
        return paramTypes;
    }

    void setVisitorState(VisitorState state) {
        this.state = state;
        classTypeVisitor.state = state;
    }

    @Override
    protected AnnotatedClassType returnType(ExecutableElement method) {
        AnnotatedClassType returnType = super.returnType(method);
        Set<AnnotationLocation> locs = immutableFinder.visit(method.getReturnType(), null);
        for (AnnotationLocation loc : locs) {
            returnType.includeAt(Immutable.class, loc);
        }
        if (method.getKind() != ElementKind.CONSTRUCTOR &&
                IGJImmutability.getIGJAnnotationDataAt(returnType, AnnotationLocation.RAW) == null)
            returnType.include(Mutable.class);
        return returnType;
    }
    
    private Map<String, AnnotationData> getBinding(AnnotatedClassType type) {
        AnnotationData annotation = IGJImmutability.getIGJAnnotationDataAt(type, AnnotationLocation.RAW);
        
        TypeMirror t;
        if (type.getTree() != null) {
            t = trees.getTypeMirror(TreePath.getPath(root, type.getTree()));
        } else if (type.getElement() != null) {
            t = type.getElement().asType();
        } else
            return Collections.emptyMap();
        
        if (t.getKind() != TypeKind.DECLARED)
            return Collections.emptyMap();
        
        DeclaredType dt = (DeclaredType)t;
        AnnotatedClassType classType = getClass(dt.asElement());
        if (classType.hasAnnotationAt(I.class, AnnotationLocation.RAW))
            return Collections.singletonMap(
                    IGJImmutability.getID(
                            IGJImmutability.getIGJAnnotationDataAt(classType,
                                    AnnotationLocation.RAW)),
                    annotation);

        return Collections.emptyMap();
    }
    
    @Override
    protected AnnotatedClassType receiverType(ExecutableElement method) {
        AnnotatedClassType receiverType = super.receiverType(method);

        
        if (IGJImmutability.getIGJAnnotationDataAt(receiverType,
                AnnotationLocation.RAW) != null) 
            return receiverType;

        IGJImmutability classImmutability =
            IGJImmutability.getIGJImmutabilityAt(
                    getClass(ElementUtils.enclosingClass(method)),
                    AnnotationLocation.RAW);

        if (classImmutability == IGJImmutability.IMMUTABLE) {
            // Include ReadOnly instead of Immutable to ease inference
            // Notice: Annotate
            receiverType.include(annotations.createAnnotation(
                    ReadOnly.class.getCanonicalName(),
                    AnnotationLocation.RAW));
        } else if (classImmutability == IGJImmutability.MUTABLE) {
            receiverType.include(Mutable.class);
        } else if (method.getKind() == ElementKind.CONSTRUCTOR) {
                receiverType.include(DEFAULT_CONSTRUCTOR_RECEIVER_ANNOTATION);
        } else {
            receiverType.include(DEFAULT_METHOD_RECEIVER_ANNOTATION);
        }

        return receiverType;
    }
    
    /**
     * Finds the union of the two {@code AnnotationData} sets of IGJ
     * immutability annotations ({@code ReadOnly}, {@code Immutable},
     * and {@code Mutable}).
     * 
     * If a location has multiple annotations, assigns a
     * {@code ReadOnly} annotation to the location.
     * 
     * @param c1
     *            a {@code Set} of IGJ {@code AnnotationData}s
     * @param c2
     *            a {@code Set} of IGJ {@code AnnotationData}s
     * @return the union of the two {@code Set}s
     */
    protected Set<AnnotationData> union(Collection<AnnotationData> c1,
            Collection<AnnotationData> c2) {
        Map<AnnotationLocation, AnnotationData> union =
            new HashMap<AnnotationLocation, AnnotationData>();

        for (AnnotationData annon : c1)
            union.put(annon.getLocation(), annon);

        for (AnnotationData annon : c2) {
            AnnotationLocation location = annon.getLocation();
            AnnotationData oldAnnon = union.put(location, annon);
            if (oldAnnon != null && !IGJImmutability.isEqual(oldAnnon, annon)) {
                union.put(location, annotations.createAnnotation(
                        ReadOnly.class.getCanonicalName(), location));
            }
        }

        Set<AnnotationData> results =
            new HashSet<AnnotationData>(union.values());
        return results;
    }

    /**
     * Resolves the binding for the passed methodType, by replacing
     * the wildcard annotations with the corresponding annotations
     * using the bindings.
     * 
     * @param methodType
     *            the method type to be resolved
     * @param bindings
     *            the wildcard ID's mapped to the corresponding
     *            annotation
     * @return true if any wildcard has been replaced/resolved.
     */
    protected boolean resolveBindings(AnnotatedMethodType methodType,
            Map<String, AnnotationData> bindings) {
        if (methodType == null || bindings == null)
            return false;
        boolean result = false;

        // TODO: Clone method
        AnnotatedMethodType method = methodType;

        // No need to resolve receiver! Yay! Less Coding!
        // IGJ prohibits Threw Annotation

        resolveBindings(method.getAnnotatedReturnType(), bindings);
        for (AnnotatedClassType type : method.getAnnotatedParameterTypes()) {
            // Be aware of order
            result = resolveBindings(type, bindings) || result;
        }

        return result;
    }

    /**
     * Resolves the wildcards in type, based on the mappings in
     * bindings.
     * 
     * When Called on
     * 
     * @W(0) Map<@W(1) Date, W(2) Calendar>, with the following
     *       bindings: 1 = {@code @Immutable } 2 = {@code @ReadOnly }
     * 
     * it'll mutate type to
     * @W(0) Map<@Immutable Date,
     * @ReadOnly Calendar>
     * 
     * @param type
     * @param bindings
     * @return true iff any wildcard has been resolved
     */
    protected boolean resolveBindings(AnnotatedClassType type,
            Map<String, AnnotationData> bindings) {
        if (type == null || bindings == null)
            return false;

        boolean result = false;
        for (AnnotationData annotation : type
                .getAnnotationData(I.class, true)) {
            String value = IGJImmutability.getID(annotation);
            AnnotationData resolvedAnnotation = bindings.get(value);
            if (resolvedAnnotation == null)
                continue; // We cannot get rid of this!

            AnnotationData newAnnotation =
                annotations.createAnnotation(IGJImmutability
                        .getAnnotationTypeName(resolvedAnnotation),
                        annotation.getLocation(), resolvedAnnotation
                                .getValues());
            type.exclude(annotation);
            type.include(newAnnotation);
            result = true;
        }
        return result;
    }

    /**
     * Resolves the binding for a method based on the arguments that
     * are based on the arguments that are passed to it. It resolves
     * the wildcard to the most restrictive annotation.
     * 
     * It replaces all unresolved wildcards with an
     * {@code IGJPlaceHolder}.
     * 
     * @param method
     *            the method to be resolved
     * @param arguments
     *            {@code AnnotatedClassType}'s of the arguments
     *            passed to the method in the invocation in the same
     *            order
     * @return true iff any wildcard is resolved
     */
    protected boolean resolveBindings(AnnotatedMethodType method,
            List<AnnotatedClassType> arguments) {
        assert method.getAnnotatedParameterTypes().size() == arguments.size();

        // First Pass : Figure out bindings
        Map<String, AnnotationData> bindings =
            new HashMap<String, AnnotationData>();
        List<AnnotatedClassType> requiredArgs =
            method.getAnnotatedParameterTypes();
        for (int i = 0; i < requiredArgs.size(); ++i) {
            AnnotatedClassType requiredArg = requiredArgs.get(i);
            AnnotatedClassType actualArg = arguments.get(i);

            for (AnnotationData wildcard : requiredArg.getAnnotationData(
                    I.class, true)) {
                AnnotationData actualAnnotation =
                    IGJImmutability.getIGJAnnotationDataAt(actualArg,
                            wildcard.getLocation());
                String id = IGJImmutability.getID(wildcard);
                AnnotationData previous = bindings.put(id, actualAnnotation);
                if (previous != null
                        && !IGJImmutability.isSubtype(actualAnnotation,
                                previous)) {
                    bindings.put(id, annotations.createAnnotation(
                            ReadOnly.class.getCanonicalName(),
                            AnnotationLocation.RAW, null));
                }
            }
        }

        // SecondPass: resolve
        boolean result = false;
        for (AnnotatedClassType requiredArg : requiredArgs) {
            // Watch for order
            result = resolveBindings(requiredArg, bindings) || result;
        }
        // Watch for order
        AnnotatedClassType returnType = method.getAnnotatedReturnType();
        result = resolveBindings(returnType, bindings) || result;

        // replace any left wildcards with placement holder
        for (AnnotationData wildcard : returnType.getAnnotationData(I.class,
                true)) {
            returnType.exclude(wildcard);
            returnType.include(annotations.createAnnotation(
                    IGJPlaceHolder.class.getCanonicalName(), wildcard
                            .getLocation()));
        }

        return result;
    }
    
    /*
     * This exists to fix a bug in the Checkers framework. Need to determine
     * what the expected behavior is
     */
    @Override
    protected void aliasFromElement(Element elt, AnnotatedClassType type) {
        if (elt == null)
            return;

        Element te = env.getTypeUtils().asElement(elt.asType());
        if (te != null) {
            for (AnnotationMirror mirror : te.getAnnotationMirrors()) {
                InternalAnnotation anno = annotations.createAnnotation(mirror);
                if (anno.getTarget().type == TargetType.CLASS_EXTENDS || anno.getTarget().type == TargetType.CLASS_EXTENDS_GENERIC_OR_ARRAY) {
                    if (IGJImmutability.isIGJAnnotation(anno) && IGJImmutability.getIGJImmutabilityAt(type, anno.getLocation()) != null)
                        continue;
                    type.annotate(anno);
                }
            }
        }

    }

    /**
     * A location visitor that finds the location of the immutable
     * types by default.
     * 
     * The visitor returns a set of the primitives (or their boxed
     * types) and String types.
     */
    private TypeVisitor<Set<AnnotationLocation>, Void> immutableFinder =
        new LocationVisitor<Set<AnnotationLocation>, Void>() {

            @Override
            protected Set<AnnotationLocation> reduce(
                    Set<AnnotationLocation> s1, Set<AnnotationLocation> s2) {
                Set<AnnotationLocation> result =
                    new HashSet<AnnotationLocation>();
                if (s1 != null)
                    result.addAll(s1);
                if (s2 != null)
                    result.addAll(s2);
                return result;
            }

            protected Set<AnnotationLocation> defaultAction(TypeMirror e,
                    Void p) {
                return Collections.<AnnotationLocation> emptySet();
            }

            /*
             * If the "deep" component of the array is immutable, it adds to
             * the return list the location of the last level.
             * 
             * Note that this only work if we have the type of the entire
             * array.
             * 
             * TODO: Todo check for partial array type.
             */
            public Set<AnnotationLocation> visitArray(ArrayType t, Void p) {
                // TODO: Test this
                // Assume that we have the entire array!

                // Find the dimension of the array
                // and whether it is an array of String or primitives
                TypeMirror component = TypesUtils.getDeepComponent(t);
                int dimensions = TypesUtils.getArrayDimensions(t);
                int locationOffset = 
                    (component.getKind() != TypeKind.DECLARED) ? 0 :
                        ((DeclaredType)component).getTypeArguments().size();
                Set<AnnotationLocation> results =
                    new HashSet<AnnotationLocation>();
                
                if (component.getKind() == TypeKind.DECLARED) {
                    for (AnnotationLocation loc : scan(component, p)) {
                        if (!loc.equals(AnnotationLocation.RAW))
                            results.add(loc);
                    }
                }
                
                if (isImmutable(component, true)) {
                    List<Integer> newLoc =
                        new ArrayList<Integer>(getCurrentLocation().asList());
                    newLoc.add(locationOffset + dimensions - 1);
                    results =  reduce(results, Collections.singleton(
                            AnnotationLocation.fromList(newLoc)));
                }
                
                
                return results;
            }

            public Set<AnnotationLocation> visitDeclared(DeclaredType t,
                    Void p) {
                // Need to take care of String Literals too!
                Set<AnnotationLocation> result =
                    Collections.<AnnotationLocation> emptySet();
                
                // is a Lateral
                if (isImmutable(t, true))
                    result = Collections.singleton(getCurrentLocation());
//                else if (!p.contains(t.asElement()) 
//                    && IGJAnnotatedTypeFactory.this.getClass(
//                            t.asElement()).hasAnnotationAt(Immutable.class,
//                                    AnnotationLocation.RAW))
//                    result = Collections.singleton(getCurrentLocation());

                return reduce(result, super.visitDeclared(t, p));
            }

            @Override
            public Set<AnnotationLocation> visitPrimitive(PrimitiveType t, Void p) {
                return Collections.singleton(getCurrentLocation());
            }
            
            public Set<AnnotationLocation> visitTypeVariable(TypeVariable t,
                    Void p) {
                return scan(t.getUpperBound(), p);
            }

            public Set<AnnotationLocation> visitWildcard(WildcardType t,
                    Void p) {
                // Not quite right
                if (t.getSuperBound() == null || t.getSuperBound().getKind() == TypeKind.TYPEVAR)
                    return Collections.emptySet();
                else
                    return scan(t.getSuperBound(), p);
            }
        };
        
        @Deprecated
        private boolean isBoxedPrimitive(TypeMirror t) {
            // worry about boxing!
            // TODO: Remove this once the JDK is annotated
            try {
                return (env.getTypeUtils().unboxedType(t) != null);
            } catch (IllegalArgumentException e) {
                return false;
            }
        }
        
        protected boolean isImmutable(TypeMirror t, boolean declared) {
            final boolean isPrimitive = t.getKind().isPrimitive();
            final boolean isString = 
                (t.getKind() == TypeKind.DECLARED) &&
                        TypesUtils.getQualifiedName((DeclaredType)t).contentEquals("java.lang.String");
            final boolean isDeclaredImmutable =
                declared && (t.getKind() == TypeKind.DECLARED) &&
                    getClass(((DeclaredType)t).asElement())
                    .hasAnnotationAt(Immutable.class, AnnotationLocation.RAW);
            final boolean isBoxedPrimitive = isBoxedPrimitive(t);

            return (isPrimitive || isString || isBoxedPrimitive || isDeclaredImmutable);
        }
        
        protected Map<String, AnnotationData> getThisBindings(AnnotatedClassType cls) {
            if (!cls.hasAnnotationAt(I.class, AnnotationLocation.RAW) ||
                    state.thisImmutability == IGJImmutability.ASSIGNSFIELDS ||
                    state.thisImmutability == IGJImmutability.READONLY)
                return Collections.emptyMap();
            
            String i = IGJImmutability.getID(
                    cls.getAnnotationData(I.class, true).iterator().next());
            AnnotationData annon =
                annotations.createAnnotation(
                        state.thisImmutability.getAnnotation().getCanonicalName(), AnnotationLocation.RAW);
            
            return Collections.singletonMap(i, annon);
        }


    /**
     * 
     * A helper class that returns the type that corresponds to the tree.
     */
    private static class ClassTypeVisitor extends
            SimpleTreeVisitor<AnnotatedClassType, AnnotatedClassType> {

        IGJAnnotatedTypeFactory factory;
        VisitorState state;

        public ClassTypeVisitor(IGJAnnotatedTypeFactory factory) {
            this.factory = factory;
            this.state = factory.state;
        }

        @Override
        protected AnnotatedClassType defaultAction(Tree tree,
                AnnotatedClassType type) {
            // Default action. Simply return the same passed type
            return type;
        }

        @Override
        public AnnotatedClassType visitNewClass(NewClassTree tree,
                AnnotatedClassType type) {
            if (IGJImmutability.getIGJAnnotationDataAt(type, AnnotationLocation.RAW) == null)
                return factory.getConstructor(tree).getAnnotatedReturnType();
            return type;
        }

        @Override
        public AnnotatedClassType visitIdentifier(IdentifierTree tree,
                AnnotatedClassType type) {

            if (tree.getName().contentEquals("this") ||
                    tree.getName().contentEquals("super")) {
                if ((state.thisImmutability == IGJImmutability.MUTABLE)
                        || state.thisImmutability == IGJImmutability.IMMUTABLE) {
                    type.include(state.thisImmutability.getAnnotation());
                    return type;
                }
                
                // Handle self references!
                // Self reference inheret the immutability of the enclosing scope
                assert ((state.thisImmutability == IGJImmutability.ASSIGNSFIELDS)
                        || (state.thisImmutability == IGJImmutability.READONLY));
                // If class is declared with @I, 'this' reference has @I
                // immutability too within ReadOnly or AssignsFields method.
                //
                // This is important for escaping this.
                ClassTree enclosingClass = TreeUtils.enclosingClass(
                        TreePath.getPath(factory.root, tree));
                AnnotatedClassType classType = factory.getClass(enclosingClass);

                if (classType.hasAnnotationAt(I.class, AnnotationLocation.RAW) ||
                        classType.hasAnnotationAt(Immutable.class, AnnotationLocation.RAW))
                    // Notice the annotate not include
                    type.annotate(IGJImmutability.getIGJAnnotationDataAt(
                            classType, AnnotationLocation.RAW));
                else
                    type.include(ReadOnly.class);
            } else if (type.getElement().getKind() == ElementKind.FIELD) {
                // Trick does not necessary work
                factory.resolveBindings(type, factory.getThisBindings(
                        factory.getClass(ElementUtils.enclosingClass(type.getElement()))));
            }

            return type;            
        }

        /*
         * Binary trees has primitive type, thus immutable by default
         */
        @Override
        public AnnotatedClassType visitBinary(BinaryTree tree,
                AnnotatedClassType type) {
            type.include(Immutable.class);
            return type;
        }

        /*
         * InstanceOf trees return boolean
         */
        @Override
        public AnnotatedClassType visitInstanceOf(InstanceOfTree tree,
                AnnotatedClassType type) {
            type.include(Immutable.class);
            return type;
        }

        @Override
        public AnnotatedClassType visitLiteral(LiteralTree tree,
                AnnotatedClassType type) {
            if (tree.getKind() == Tree.Kind.NULL_LITERAL)
                // 'null' could be of any immutable type
                type.include(IGJPlaceHolder.class);
            else
                // Other literals are primitive or string, thus immutable
                type.include(Immutable.class);
            return type;
        }

        @Override
        public AnnotatedClassType visitUnary(UnaryTree tree,
                AnnotatedClassType type) {
            // Unary trees have primitive value
            type.include(Immutable.class);
            return type;
        }

        @Override
        public AnnotatedClassType visitPrimitiveType(PrimitiveTypeTree tree,
                AnnotatedClassType type) {
            type.include(Immutable.class);
            return type;
        }

        @Override
        public AnnotatedClassType visitVariable(VariableTree tree,
                AnnotatedClassType type) {
            // Disable the following, due to a bug in handling parametrized types
//            AnnotatedClassType varType = factory.getClass(tree.getType());
//            if (varType.getElement() == null)
//                // FIXME: Should return an element for primitive
//                // types!
//                type.include(Immutable.class);
            return type;
        }

        @Override
        public AnnotatedClassType visitConditionalExpression(
                ConditionalExpressionTree tree, AnnotatedClassType type) {
            // For Conditional expression, find the union of type of true
            // and false expressions
            Set<AnnotationData> trueAnnotations = 
                IGJImmutability.getIGJAnnotationData(
                        factory.getClass(tree.getTrueExpression()));
            
            Set<AnnotationData> falseAnnotations =
                IGJImmutability.getIGJAnnotationData(
                        factory.getClass(tree.getFalseExpression()));
            
            Set<AnnotationData> union =
                factory.union(trueAnnotations, falseAnnotations);
            
            for (AnnotationData annon : union)
                type.include(annon);
            
            return type;
        }

        @Override
        public AnnotatedClassType visitClass(ClassTree tree,
                AnnotatedClassType type) {
            if (IGJImmutability.getIGJAnnotationDataAt(type,
                    AnnotationLocation.RAW) != null) 
                // Already has immutability type! done!
                return type;
            
            TreePath path = TreePath.getPath(factory.root, tree);
            if (TreeUtils.enclosingMethod(path) != null) {
                // Inner class within a method!
                AnnotatedClassType receiverType =
                    factory.getMethod(TreeUtils.enclosingMethod(path))
                    .getAnnotatedReceiverType();
                AnnotationData annotation =
                    IGJImmutability.getIGJAnnotationDataAt(receiverType,
                            AnnotationLocation.RAW);
                type.include(annotation);
            } else {
                // TODO: Handle inner static classes if needed!
                type.include(DEFAULT_CLASS_ANNOTATION);
            }
            
            return type;
        }

        @Override
        public AnnotatedClassType visitMethodInvocation(
                MethodInvocationTree tree, AnnotatedClassType type) {
            AnnotatedMethodType method = factory.getMethod(tree);
            return method.getAnnotatedReturnType();
        }

        @Override
        public AnnotatedClassType visitMemberSelect(MemberSelectTree tree,
                AnnotatedClassType type) {
            if (ElementUtils.isStatic(type.getElement()))
                return type;

            // Try to resolve any wildcard annotation variable
            if (!type.getAnnotationData(I.class.getCanonicalName(), true)
                    .isEmpty()) {
                MemberSelectTree selectTree = (MemberSelectTree) tree;
                AnnotatedClassType enclosingType =
                    factory.getClass(selectTree.getExpression());
                factory.resolveBindings(type, 
                        factory.getBinding(enclosingType));
            }

            return type;
        }
    }

}
